<?php
/*************************

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

**********************************************/

chdir('../../');

// Include OAuth classes
require_once 'plugins/oauth/store/oauth_store_mysql.php';
require_once 'plugins/oauth/oauth_server.php';

define('IN_COPPERMINE', true);

// Inspekt
require_once('include/inspekt.php');

$superCage = Inspekt::makeSuperCage();

if ($superCage->post->getAlpha('function')){
  require_once 'include/config.inc.php';
} else {
  require_once 'include/init.inc.php';
}

define('FIRST_USER_CAT', 10000);

$CONFIG['TABLE_PICTURES']      = $CONFIG['TABLE_PREFIX'].'pictures';
$CONFIG['TABLE_ALBUMS']        = $CONFIG['TABLE_PREFIX'].'albums';
$CONFIG['TABLE_COMMENTS']      = $CONFIG['TABLE_PREFIX'].'comments';
$CONFIG['TABLE_CATEGORIES']    = $CONFIG['TABLE_PREFIX'].'categories';
$CONFIG['TABLE_CONFIG']        = $CONFIG['TABLE_PREFIX'].'config';
$CONFIG['TABLE_USERGROUPS']    = $CONFIG['TABLE_PREFIX'].'usergroups';
$CONFIG['TABLE_VOTES']         = $CONFIG['TABLE_PREFIX'].'votes';
$CONFIG['TABLE_USERS']         = $CONFIG['TABLE_PREFIX'].'users';
$CONFIG['TABLE_BANNED']        = $CONFIG['TABLE_PREFIX'].'banned';
$CONFIG['TABLE_EXIF']          = $CONFIG['TABLE_PREFIX'].'exif';
$CONFIG['TABLE_FILETYPES']     = $CONFIG['TABLE_PREFIX'].'filetypes';
$CONFIG['TABLE_ECARDS']        = $CONFIG['TABLE_PREFIX'].'ecards';
$CONFIG['TABLE_FAVPICS']       = $CONFIG['TABLE_PREFIX'].'favpics';
$CONFIG['TABLE_BRIDGE']        = $CONFIG['TABLE_PREFIX'].'bridge';
$CONFIG['TABLE_VOTE_STATS']    = $CONFIG['TABLE_PREFIX'].'vote_stats';
$CONFIG['TABLE_HIT_STATS']     = $CONFIG['TABLE_PREFIX'].'hit_stats';
$CONFIG['TABLE_TEMP_MESSAGES'] = $CONFIG['TABLE_PREFIX'].'temp_messages';
$CONFIG['TABLE_CATMAP']        = $CONFIG['TABLE_PREFIX'].'categorymap';
$CONFIG['TABLE_LANGUAGE']      = $CONFIG['TABLE_PREFIX'].'languages';
$CONFIG['TABLE_DICT']          = $CONFIG['TABLE_PREFIX'].'dict';

$options['server'] = $CONFIG['dbserver'];
$options['username'] = $CONFIG['dbuser'];
$options['password'] = $CONFIG['dbpass'];
$options['database'] = $CONFIG['dbname'];
$oauth_store = OAuthStore::instance('mysql', $options);

$token = $superCage->post->getAlnum('oauth_token');
if (!$token) {
    $token = $superCage->get->getAlnum('oauth_token');
}

if ($token) {
    $user_id = $oauth_store->lookup_id($token);
    if ($user_id) {
        define('USER_ID', $user_id);
        define('USER_NAME', 'API');

        $group_info = $oauth_store->group_info($user_id);
        if ($group_info['can_upload']) {
            define('USER_CAN_UPLOAD_PICTURES', true);
        }
        if ($group_info['admin_access']) {
            define('GALLERY_ADMIN_MODE', true);
            define('USER_IS_ADMIN', true);
        }
    }
}

// End main code

/** 
 * Object to process OAuth requests for Coppermine
 */
class cpgOAuth extends OAuthServer {
    function __construct($consumer_key, $nonce, $timestamp, $signature_method, $signature, $token) {

        $this->setParam('oauth_token', $token, true);
        $this->setParam('oauth_consumer_key', $consumer_key, true);
        $this->setParam('oauth_nonce', $nonce, true);       
        $this->setParam('oauth_timestamp', $timestamp, true);
        $this->setParam('oauth_signature_method', $signature_method, true);
        $this->setParam('oauth_signature', $signature, true);
        parent::__construct();
    }

    public function access_protected_resource() {
        global $CONFIG, $THEME_DIR, $USER, $CAT_LIST;
        global  $cpg_udb; // Needed for "lastcomby" meta album in picture list
        try {
            $result = $this->verify('access');

            if ($result != null) {
                define('API_CALL', true);
                $superCage = Inspekt::makeSuperCage();
                $matches = $superCage->post->getMatched('function', '/^[a-z]+$/');
                if (count($matches[0]) > 0 && !function_exists('cpgGetMicroTime')){
                    //Very ugly
                    function cpgGetMicroTime()
                    {
                        list($usec, $sec) = explode(' ', microtime());
                        return ((float)$usec + (float)$sec);
                    }
                }
                switch ($matches[0]) {
                case 'alblist':
                    define('IN_COPPERMINE', true);
                    require_once 'include/functions.inc.php';
                    $CONFIG['LINK_ID'] = cpg_db_connect();
                    if (!$CONFIG['LINK_ID']) {
                        log_write("Unable to connect to database: " . mysql_error(), CPG_DATABASE_LOG);
                        die('<strong>Coppermine critical error</strong>:<br />Unable to connect to database !<br /><br />MySQL said: <strong>' . mysql_error() . '</strong>');
                    }

                    $private = false;
                    if ($superCage->post->getInt('private') == 1){
                        $private = true;
                    }

                    oauth_user_albums(USER_ID, $private);
                    oauth_list_items('', '');
                    break;                           
                case 'piclist':
                    define('IN_COPPERMINE', true);
                    require_once 'include/functions.inc.php';
                    $CONFIG['LINK_ID'] = cpg_db_connect();
                    if (!$CONFIG['LINK_ID']) {
                        log_write("Unable to connect to database: " . mysql_error(), CPG_DATABASE_LOG);
                        die('<strong>Coppermine critical error</strong>:<br />Unable to connect to database !<br /><br />MySQL said: <strong>' . mysql_error() . '</strong>');
                    }

                    $private = false;
                    if ($superCage->post->getInt('private') == 1){
                        $private = true;
                    }

                    if ($superCage->post->getInt('album')) {
                        oauth_user_albums(USER_ID, $private);
                        oauth_list_items('', '');
                    } else if ($album = $superCage->post->getAlpha('album')) {
                        $allowed = array('lastcom', 'lastcomby', 'lastup', 'lastupby', 'topn', 'toprated', 'lasthits');                                 
                        if (!in_array($album, $allowed)) {
                            new OAuthException("Valid meta album names for this function are: 'lastcom', 'lastcomby', 'lastup', 'lastupby', 'topn', 'toprated', and 'lasthits'");
                        }
                        $USER['uid'] = USER_ID;
                        require 'thumbnails.php';
                    }
                    else { // No album provided
                        new OAuthException('No album provided via HTTP POST');
                    }
                    break;
                case 'commentslist':
                    define('IN_COPPERMINE', true);
                    if (!$superCage->post->getInt('photo')) {
                        new OAuthException('No photo provided via HTTP POST');
                        break;
                    }

                    require_once 'include/functions.inc.php';
                    $CONFIG['LINK_ID'] = cpg_db_connect();
                    if (!$CONFIG['LINK_ID']) {
                        log_write("Unable to connect to database: " . mysql_error(), CPG_DATABASE_LOG);
                        die('<strong>Coppermine critical error</strong>:<br />Unable to connect to database !<br /><br />MySQL said: <strong>' . mysql_error() . '</strong>');
                    }

                    $private = false;
                    if ($superCage->post->getInt('private') == 1){
                        $private = true;
                    }

                    if ($superCage->post->getInt('album')) {
                        oauth_user_albums(USER_ID, $private);
                        oauth_list_items('', '');
                    } else if ($album = $superCage->post->getAlpha('album')) {
                        $allowed = array('lastcom', 'lastcomby', 'lastup', 'lastupby', 'topn', 'toprated', 'lasthits');                                 
                        if (!in_array($album, $allowed)) {
                            new OAuthException("Valid meta album names for this function are: 'lastcom', 'lastcomby', 'lastup', 'lastupby', 'topn', 'toprated', and 'lasthits'");
                        }
                        $USER['uid'] = USER_ID;
                        die('TODO:');
                    }
                    else { // No album provided
                        new OAuthException('No album provided via HTTP POST');
                    }
                    break;
                case 'search':
                    define('IN_COPPERMINE', true);
                    require 'include/init.inc.php';
                    require 'thumbnails.php';
                    break;
                case 'catlist':
                    define('IN_COPPERMINE', true);
                    require 'include/init.inc.php';                            
                    oauth_cat_list();
                    break;
                default:
                    throw new OAuthException('No function specified via HTTP POST');
                }
            }
        }

        catch (OAuthException $e) {
            header('HTTP/1.1 401 Access Denied');
            header('Content-Type: text/xml');
            throw new OAuthException($e->getMessage());
        } catch (Exception $e) {
            die("exception: ".$e->getMessage());
        }
        OAuthRequestLogger::flush();
        exit();
    }
}

function oauth_api_message($message) {
    die("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n" . '<api_message>' . $message . '</api_message>');   
}

/**
 * function oauth_cat_list()
 *
 * Geta hierarchical list of categories, in XML
 *
 */
function oauth_cat_list() {
    global $CONFIG, $CAT_LIST;
    $superCage = Inspekt::makeSuperCage();
    $cat = 0;
    if ($superCage->post->getInt('catid')) {
        $cat = $superCage->post->getInt('catid'); 
    }
    oauth_subcategory_data($cat);
    echo "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n";
    echo "<cat_list>\n";     
    $catlevels = array();                       
    // Loop through the category list and store the parent of each one in an array
    foreach ($CAT_LIST as $category) {
        // oauth_subcategory_data() prints '&nbsp;' three times for each level of depth of categories
        $catlevels[] = substr_count($category['name'], 'nbsp;') / 3;
    }

    // Loop through again, printing out the list of categories
    $count = 0; // The iteration of the loop
    $level = 0; // How many levels deep in the hierarchy a category is
    $indent2 = ''; // The spacing for closing tags
    foreach ($CAT_LIST as $category) {
        $level = $catlevels[$count];
        $indent = '';
        for ($i = 0; $i < $level; $i++) {
            $indent .= ' ';
        }

        $previous = $catlevels[$count - 1];
        $current = $catlevels[$count];                                
        if ($previous == $current && $count > 0) {
            $indent2 = substr($indent2, 1); // 1 less leading space for each closing tag in a series
            echo $indent2 . "</category>\n";
        }

        else if ($previous > $current) {
            for ($j = 0; $j <= $previous - $current; $j++) {
                $indent2 = substr($indent2, 1);
                echo $indent2 . "</category>\n";
            }
        }

        $name = preg_replace("/((&nbsp;){3})+/", "", $category['name']); // Clean up the name text
        echo $indent . '<category id="' . $category['cid'] . "\" name=\"" . $name . "\">\n";
        $indent2 .= ' '; // 1 more leading space for each <category> element

        $count++;
    }

    // The final (set of) closing tag(s)
    if ($count > 0) {
        for ($i = 0; $i <= $level; $i++) {
            $indent2 = substr($indent2, 1);
            echo $indent2 . "</category>\n";
        }
    }

    echo "</cat_list>";
}

//$private == false -> only public
//$private == true -> only private
function oauth_user_albums($user_id, $private = true) {
    global $CONFIG, $USER_DATA, $public_albums_list, $user_albums_list;
    $superCage = Inspekt::makeSuperCage();
    // Filter albums by category
    $api_sql = '';
    $api_sql2 = '';
    if (defined('API_CALL')) {
        $catid = $superCage->post->getInt('cat');
        if ($superCage->post->getAlpha('function') == 'alblist') {
            global $CAT_LIST;
            oauth_subcategory_data($catid);
            if (count($cat) > 0){
                $api_sql .= " AND (";
                foreach ($CAT_LIST as $cat) {
                    $api_sql .= "category = {$cat['cid']} OR ";
                }
                $api_sql .= "category = $catid)";
            }
        }
        if ($superCage->post->getAlpha('function') == 'piclist' && $album = $superCage->post->getInt('album')) {
            $api_sql2 = " AND aid = $album";
        }

        if ($catid >= 0){
            $query = "SELECT aid, title, category as cid, name, user_name FROM {$CONFIG['TABLE_ALBUMS']} LEFT OUTER JOIN {$CONFIG['TABLE_CATEGORIES']} ON cid = category JOIN {$CONFIG['TABLE_USERS']} ON user_id = owner WHERE category != 0 AND ((visibility = '0' OR visibility IN ".oauth_user_groups_set($user_id).") OR (owner=".$user_id." AND visibility !=".(FIRST_USER_CAT + $user_id).")){$api_sql2} AND category = " . $catid;
        } else {
            $query = "SELECT aid, title, category as cid, name, user_name FROM {$CONFIG['TABLE_ALBUMS']} LEFT OUTER JOIN {$CONFIG['TABLE_CATEGORIES']} ON cid = category JOIN {$CONFIG['TABLE_USERS']} ON user_id = owner WHERE category != 0 AND ((visibility = '0' OR visibility IN ".oauth_user_groups_set($user_id).") OR (owner=".$user_id." AND visibility !=".(FIRST_USER_CAT + $user_id).")){$api_sql2} ". $api_sql;
        }
    } else {
        die('Error');
    }

    if ($private == false){
        $public_albums = cpg_db_query($query);
        //select albums that don't belong to a category

        $query = "SELECT aid, title, category as cid, user_name FROM {$CONFIG['TABLE_ALBUMS']} JOIN {$CONFIG['TABLE_USERS']} ON user_id = owner WHERE category = 0 AND ((visibility = '0' OR visibility IN ".oauth_user_groups_set($user_id).") OR (owner=".$user_id.")){$api_sql2}";
        $public_albums_no_cat = cpg_db_query($query);
        if (mysql_num_rows($public_albums)) {
            $public_albums_list = cpg_db_fetch_rowset($public_albums);
        } else {
            $public_albums_list = array();
        }

        //do the same for non categorized albums
        if (mysql_num_rows($public_albums_no_cat)) {
            $public_albums_list_no_cat = cpg_db_fetch_rowset($public_albums_no_cat);
        } else {
            $public_albums_list_no_cat = array();
        }
    } else {
        $public_albums_list = array();
        $public_albums_list_no_cat = array();
    }

    //merge the 2 album arrays
    $public_albums_list = array_merge($public_albums_list, $public_albums_list_no_cat);

    if ($user_id && $private) {
        if ($api_sql) {
            $api_sql = preg_replace('/AND/', '', $api_sql);
            $api_sql .= " AND ";
        }
        if ($api_sql2) {
            $api_sql = "aid=$album AND ";
        }
        if ($catid >= 0){
            $query = "SELECT aid, title, category as cid, user_name FROM {$CONFIG['TABLE_ALBUMS']} JOIN {$CONFIG['TABLE_USERS']} ON user_id = owner WHERE category = " . $catid . " AND visibility='" . (FIRST_USER_CAT + $user_id) . "' ORDER BY title";
        } else {
            $query = "SELECT aid, title, category as cid, user_name FROM {$CONFIG['TABLE_ALBUMS']} JOIN {$CONFIG['TABLE_USERS']} ON user_id = owner WHERE " . $api_sql . " visibility='" . (FIRST_USER_CAT + $user_id) . "' ORDER BY title";
        }
        $user_albums = cpg_db_query($query);

        if (mysql_num_rows($user_albums)) {
            $user_albums_list = cpg_db_fetch_rowset($user_albums);
        } else {
            $user_albums_list = array();
        }

    } else {
        $user_albums_list = array();
    }

    if (count($user_albums_list) + count($public_albums_list) == 0){
        die ('no albums');
    }
}

function oauth_user_groups_set($user_id){
    global $CONFIG;
    $sql = "SELECT user_id, user_group "
            . "FROM ".$CONFIG['TABLE_USERS']." "
            . "WHERE user_id=".$user_id;
    //$sql = "SELECT u.user_id AS id, ug.group_id AS group_id "
    //            . "FROM $CONFIG['TABLE_USERS'] AS u, $CONFIG['TABLE_USERGROUPS'] AS ug "
    //            . "WHERE u.user_id=ug.user_id AND u.='".$user_id."'";

    $result = cpg_db_query($sql);

    if (mysql_num_rows($result)){
        $row = mysql_fetch_assoc($result);
        mysql_free_result($result);
        $group = array(0 => $row['user_group']);
        if ($group[0] == 1){
            $group[1] = 2;
        }
        $user_group_set = '(' . implode(',', $group) . ')';

        return $user_group_set;
    }

    return '()';
}

function oauth_list_items($text, $name, $type=0) {
    $superCage = Inspekt::makeSuperCage();
    // Pull the $CONFIG array and the GET array into the function
    global $CONFIG;

    // Also pull the album lists into the function
    global $user_albums_list, $public_albums_list;

    // Check to see if an album has been preselected by URL addition or the last selected album. If so, make $sel_album the album number. Otherwise, make $sel_album 0.
    if ($superCage->get->keyExists('album')) {
        $sel_album = $superCage->get->getInt('album');
    } elseif ($superCage->post->keyExists('album')) {
        $sel_album = $superCage->post->getInt('album');
    } else {
        $sel_album = 0;
    }

    // Get the ancestry of the categories
    $vQuery = "SELECT cid, parent, name FROM " . $CONFIG['TABLE_CATEGORIES'] . " WHERE 1";
    $vResult = cpg_db_query($vQuery);
    $vRes = cpg_db_fetch_rowset($vResult);
    mysql_free_result($vResult);
    foreach ($vRes as $vResI => $vResV) {
        $vResRow = $vRes[$vResI];
        $catParent[$vResRow['cid']] = $vResRow['parent'];
        $catName[$vResRow['cid']] = $vResRow['name'];
    }
    $catAnces = array();
    foreach ($catParent as $cid => $cid_parent) {
        $catAnces[$cid] = '';
        while ($cid_parent != 0) {
            $catAnces[$cid] = $catName[$cid_parent] . ($catAnces[$cid]?' - '.$catAnces[$cid]:'');
            $cid_parent = $catParent[$cid_parent];
        }
    }

    // Reset counter
    $list_count = 0;

    // Cycle through the User albums
    foreach($user_albums_list as $album) {

        // Add to multi-dim array for later sorting
        $listArray[$list_count]['cat'] = 'personal_albums';
        $listArray[$list_count]['aid'] = $album['aid'];
        $listArray[$list_count]['title'] = $album['title'];
        $listArray[$list_count]['user_name'] = $album['user_name'];
        $list_count++;
    }

    // Cycle through the public albums
    foreach($public_albums_list as $album) {

        // Set $album_id to the actual album ID
        $album_id = $album['aid'];

        // Add to multi-dim array for sorting later
        if ($album['name']) {
            $listArray[$list_count]['cat'] = $catAnces[$album['cid']] . ($catAnces[$album['cid']]?' - ':'') . $album['name'];
            $listArray[$list_count]['cid'] = $album['cid'];
        } else if ($album['cid'] == 0){
            $listArray[$list_count]['cat'] = 'albums_no_category';
            $listArray[$list_count]['cid'] = 0;
        } else {
            $listArray[$list_count]['cat'] = '';
            $listArray[$list_count]['cid'] = $album['cid'];;
        }
        $listArray[$list_count]['aid'] = $album['aid'];
        $listArray[$list_count]['title'] = $album['title'];
        $listArray[$list_count]['user_name'] = $album['user_name'];
        $list_count++;
    }

    if (defined('API_CALL')) {
        if ($superCage->post->getAlpha('function') == 'alblist') {
            $listArray = array_csort($listArray,'aid'); // Sort the array alphabetically by aid
            $result = array();
            $result['success'] = 'true';
            $result['data'] = $listArray;
            die(json_encode($result));
        } else if ($superCage->post->getAlpha('function') == 'piclist') {
            $listArray = array_csort($listArray,'cat','title');     // alphabetically by category name
            $listAlbumArray = array();
            foreach ($listArray as $val) { 
                // Picture list
                $result = cpg_db_query("SELECT pid, title, filepath, filename, pwidth, pheight, caption FROM {$CONFIG['TABLE_PICTURES']} WHERE aid = {$val['aid']} AND approved='YES'");
                while ($row = mysql_fetch_assoc($result)) {
                    $photo = array();
                    $photo['picture_id'] = $row['pid'];
                    $photo['title'] = $row['title'];
                    $photo['file'] = "albums/".$row['filepath'] . $row['filename'];
                    $photo['caption'] = $row['caption'];
                    $text_comment = "";
                    $resultcount = cpg_db_query("SELECT COUNT(*) FROM {$CONFIG['TABLE_COMMENTS']} WHERE pid = {$photo['picture_id']} AND approval='YES'");
                    $msgcount = mysql_fetch_row($resultcount);
                    mysql_free_result($resultcount);
                    $photo['msg_count'] = $msgcount[0];
                    $resultcomment = cpg_db_query("SELECT user_name, msg_author, msg_body FROM {$CONFIG['TABLE_COMMENTS']} JOIN {$CONFIG['TABLE_USERS']} ON user_id = author_id WHERE pid = {$photo['picture_id']} AND approval='YES' ORDER BY msg_date LIMIT 1");
                    while ($rowcomment = mysql_fetch_assoc($resultcomment)) {
                        if (count($rowcomment['msg_author']) > 0){
                            $text_comment = $rowcomment['msg_author'].": ".$rowcomment['msg_body'];
                        } else {
                            $text_comment = $rowcomment['user_name'].": ".$rowcomment['msg_body'];
                        }
                    }
                    mysql_free_result($resultcomment);
                    $photo['comment'] = $text_comment;

                    $photos[] = $photo;                                           
                }
                mysql_free_result($result);
                $val['photos'] = $photos;
                $listAlbumArray[] = $val;
            }

            $result = array();
            $result['success'] = 'true';
            $result['data'] = $listAlbumArray;
            die(json_encode($result));
        } else if ($superCage->post->getAlpha('function') == 'commentslist') {
            $listArray = array_csort($listArray,'cat','title');     // alphabetically by category name
            $listCommentsArray = array();
            foreach ($listArray as $val) { 
                // Picture list
                $query = "SELECT pid FROM {$CONFIG['TABLE_PICTURES']} WHERE aid = {$val['aid']} AND approved='YES' AND pid={$superCage->post->getInt('photo')}";
                $result = cpg_db_query($query);
                while ($row = mysql_fetch_assoc($result)) {
                    $resultcomments = cpg_db_query("SELECT user_name, msg_author, msg_body, msg_date FROM {$CONFIG['TABLE_COMMENTS']} JOIN {$CONFIG['TABLE_USERS']} ON user_id = author_id WHERE pid = {$row['pid']} AND approval='YES' ORDER BY msg_date");
                    while ($rowcomments = mysql_fetch_assoc($resultcomments)) {
                        $comment = array();
                        if (count($rowcomments['msg_author']) > 0){
                            $comment['from'] = $rowcomments['msg_author'];
                        } else {
                            $comment['from'] = $rowcomments['user_name'];
                        }
                        $comment['message'] = $rowcomments['msg_body'];
                        $comment['created_time'] = $rowcomments['msg_date'];
                        $listCommentsArray[] = $comment;
                    }
                    mysql_free_result($resultcomments);
                }
                mysql_free_result($result);
            }

            $result = array();
            $result['success'] = 'true';
            $result['photo'] = "{$superCage->post->getInt('photo')}";
            $result['data'] = $listCommentsArray;
            die(json_encode($result));
        } else {
            $result = array();
            $result['success'] = 'false';
            $result['message'] = 'Unknown API function.';
            die(json_encode($result));
        }

        return;
    }
}

function oauth_subcategory_data($parent, $ident = '')
{
    global $CONFIG, $CAT_LIST;
    if ($CONFIG['categories_alpha_sort'] == 1) {
        $sort_query = 'name';
    } else {
        $sort_query = 'pos';
    }

    $sql = "SELECT cid, name, description " . "FROM {$CONFIG['TABLE_CATEGORIES']} " . "WHERE parent = '$parent' " . "ORDER BY $sort_query";
    $result = cpg_db_query($sql);
    if (($cat_count = mysql_num_rows($result)) > 0) {
        $rowset = cpg_db_fetch_rowset($result);
        $pos = 0;
        foreach ($rowset as $subcat) {
            if ($pos > 0) {
                $CAT_LIST[$subcat['cid']] = array('cid' => $subcat['cid'],
                        'parent' => $parent,
                        'pos' => $pos++,
                        'prev' => $prev_cid,
                        'cat_count' => $cat_count,
                        'name' => $ident . $subcat['name']);
                $CAT_LIST[$last_index]['next'] = $subcat['cid'];
            } else {
                $CAT_LIST[$subcat['cid']] = array('cid' => $subcat['cid'],
                        'parent' => $parent,
                        'pos' => $pos++,
                        'cat_count' => $cat_count,
                        'name' => $ident . $subcat['name']);
            }
            $last_index = $prev_cid = $subcat['cid'];
            oauth_subcategory_data($subcat['cid'], $ident . '&nbsp;&nbsp;&nbsp;'); // Note: API category list depends on these &nbsp;'s
        }
    }
}

?>